from importlib import resources

from yaml import safe_load
from jsonschema import validate


def load_schema():
    schema = resources.files(__package__).joinpath("schema.yaml").read_bytes()

    return safe_load(schema)


def validate_layout(layout):
    return validate(instance=layout, schema=load_schema())



from showy.preprocess_utils import _has_stars
from showy.preprocess_utils import _replace_stars
from showy.preprocess_utils import _find_stars_values_in_string
from showy.preprocess_utils import _replace_stars_in_string
from showy.preprocess_utils import _collect_stars_values


def test_graph_layout_stars():
    graph_layout = {
        "curves": [{"var": "Y_*_min", "legend": "*"}],
        "x_var": "time",
        "y_label": "*",
        "x_label": "Time [s]",
        "title": "*",
    }

    assert _has_stars(graph_layout) is True

    available_varnames = ["Y_CO_min", "Y_CO2_min"]
    new_graph_layouts = _replace_stars(graph_layout, available_varnames)

    assert len(new_graph_layouts) == len(available_varnames)
    for name, graph_layout_ in zip(["CO", "CO2"], new_graph_layouts):
        assert graph_layout_["curves"][0]["var"] == f"Y_{name}_min"
        assert graph_layout_["curves"][0]["legend"] == name
        assert graph_layout_["y_label"] == name
        assert graph_layout_["title"] == name


def test_find_stars_in_string():
    available_varnames = ["Y_CO_min", "Y_CO_max", "Y_CO2_min", "Y_CO2_max"]

    # one star
    var_name = "Y_*_min"
    stars_values = _find_stars_values_in_string(var_name, available_varnames)

    target_values = ["CO", "CO2"]
    assert len(stars_values) == len(target_values)

    values = [value[0] for value in stars_values]
    for target_value in target_values:
        assert target_value in values

    # two stars
    var_name = "Y_*_*"
    stars_values = _find_stars_values_in_string(var_name, available_varnames)

    target_values = [("CO", "min"), ("CO", "max"), ("CO2", "min"), ("CO2", "max")]
    assert len(stars_values) == len(target_values)
    for target_value in target_values:
        assert target_value in stars_values


def test_replace_stars_in_string():

    # one star
    var_name = "Y_*"
    new_var_name = _replace_stars_in_string(var_name, ("1",))
    assert new_var_name == "Y_1"

    # two_stars
    var_name = "Y_*_*"
    new_var_name = _replace_stars_in_string(var_name, ("1", "2"))
    assert new_var_name == "Y_1_2"


def test_collect_stars_values():
    available_varnames = ["Y_CO_min", "Y_CO_max", "Y_CO2_min", "Y_CO2_max", "Y_O_max"]

    var_names = ["Y_*_min", "Y_*_max"]
    stars_values = _collect_stars_values(var_names, available_varnames)

    target_values = ["CO", "CO2"]
    assert len(stars_values) == len(target_values)

    values = [value[0] for value in stars_values]
    for target_value in target_values:
        assert target_value in values



import pytest
from jsonschema import ValidationError

from showy.validation import validate_layout


def test_validation():
    layout = {
        "title": "SPEC",
        "graphs": [
            {
                "curves": [
                    {"var": "Y_*_max", "legend": "max"},
                    {"var": "Y_*_mean", "legend": "mean"},
                    {"var": "Y_*_min", "legend": "min"},
                ],
                "x_var": "atime",
                "y_label": "Mass fraction * [mol/m³/s]",
                "x_label": "Time [s]",
                "title": "* Mass Fraction",
            }
        ],
        "figure_dpi": 90,
        "figure_size": [14, 8],
    }

    validate_layout(layout)

    with pytest.raises(ValidationError):
        layout["titlea"] = layout["title"]
        validate_layout(layout)



import numpy as np

from showy import showy


def test_showy():
    data = dict()
    data["time"] = time = np.linspace(0, 0.1, num=256)

    for freq in np.linspace(10, 20, num=9):
        data[f"sine_{freq}"] = np.sin(time * freq * 2 * np.pi)
        data[f"cosine_{freq}"] = np.cos(time * freq * 2 * np.pi)
    data["cos"] = np.cos(time)

    # Creating a template
    template = {
        "title": "Example",
        "graphs": [
            {
                "curves": [{"var": "sine_*"}, {"var": "cosine_*"}, {"var": "cos"}],
                "x_var": "time",
                "y_label": "Sine [mol/m³/s]",
                "x_label": "Time [s]",
                "title": "Sine and cosine of frequency *",
            },
            {"curves": [{"var": "cosine_10.0"}, {"var": "sine_10.0"}], "x_var": "time"},
        ],
        "figure_structure": [3, 3],
        "figure_dpi": 92.6,
    }

    # single data
    showy(template, data, show=False)

    # multiple data
    showy(template, [data, data], show=False)



import math
import warnings

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from math import floor

from showy.preprocess_utils import preprocess_layout
from showy.validation import validate_layout


# TODO: add export of data
# TODO: layout merging
# TODO: add schema validation


def showy(layout, dataframe, show=True, preprocess_fnc=preprocess_layout):
    """It displays the desired graphs described by the provided layout
    thanks to the provided key-value object which contains
    the required data
    """
    validate_layout(layout)

    dataframes = dataframe
    if type(dataframes) not in [tuple, list]:
        dataframes = [dataframes]

    comparison_mode = len(dataframes) > 1

    available_varnames = _get_available_varnames(dataframes)

    layout = preprocess_fnc(layout, available_varnames)

    display_params = _initialize_display(layout)
    max_graphs = display_params["max_graphs"]
    fig_layout = display_params["layout"]

    figure_number = 0
    graph_position = 1
    for graph_layout in layout["graphs"]:
        if graph_position > max_graphs or figure_number == 0:
            graph_position = 1
            figure_number += 1
            figure = _build_new_figure(
                fig_number=figure_number,
                comparison_mode=comparison_mode,
                **display_params,
            )

        subplot = figure.add_subplot(*fig_layout, graph_position)
        _build_new_graph(subplot, graph_layout, dataframes)
        figure.tight_layout()
        graph_position += 1

    if show:
        plt.show()


def _get_available_varnames(dataframes):
    var_names = set(dataframes[0].keys())
    for dataframe in dataframes[1:]:
        var_names = var_names.intersection(set(dataframe.keys()))

    return list(var_names)


def _initialize_display(layout):
    """Sets some parameters needed by the function "display", thanks
    to the provided layout, and returns them in a dictionary.
    """

    n_graphs = len(layout["graphs"])

    fig_layout = layout.get("figure_structure", None)
    if fig_layout is None:
        fig_layout = [1, n_graphs] if n_graphs < 4 else [2, 3]

    n_max_graphs_fig = fig_layout[0] * fig_layout[1]
    n_figs = math.ceil(n_graphs / n_max_graphs_fig)

    display_params = {
        "layout": fig_layout,
        "max_graphs": n_max_graphs_fig,
        "n_figs": n_figs,
        "dpi": layout.get("figure_dpi", None),
        "inches_size": layout.get("figure_size", (14, 8)),
        "title": layout.get("title", ""),
    }

    return display_params


def _build_new_figure(
    dpi=None,
    inches_size=(14, 8),
    fig_number=None,
    n_figs=None,
    title="",
    comparison_mode=False,
    **kwargs,
):
    figure = plt.figure(dpi=dpi)
    figure.set_size_inches(*inches_size)

    figure_name = title
    if comparison_mode:
        figure_name += " (comparison)"
    if fig_number is not None and n_figs is not None and n_figs > 1:
        figure_name += f" ({fig_number}/{n_figs})"

    figure.canvas.manager.set_window_title(figure_name.strip())

    figure.subplots_adjust(
        left=0.11, bottom=0.12, right=0.9, top=0.9, wspace=0.38, hspace=0.32
    )

    return figure


def _build_new_graph(ax, graph_layout, dataframes):
    """Builds a new matplotlib graph in the provided subplot."""
    # TODO: handle tick label overlap
    # TODO: handle no legend space

    ax.set_xlabel(graph_layout.get("x_label", None))
    ax.set_ylabel(graph_layout.get("y_label", None))
    ax.set_title(graph_layout.get("title", None), fontweight="bold")

    for data_number, dataframe in enumerate(dataframes):
        linewidth = 2
        if data_number > 0:
            linewidth = 4
        x = dataframe[graph_layout["x_var"]]
        for curve_number, curve_layout in enumerate(graph_layout["curves"]):
            symbols_markevery = _symbols_markevery(x)
            custom_style = _custom_cycler_iterate(
                curve_number, data_number, symbols_markevery
            )
            try:
                ax.plot(
                    x,
                    dataframe[curve_layout["var"]],
                    **custom_style,
                    linewidth=linewidth,
                )
            except KeyError:
                print(
                    f"Could not find the key {curve_layout['var']} in current dataframe"
                )

    # improve legends
    handles = []

    for curve_number, curve_layout in enumerate(graph_layout["curves"]):
        label = curve_layout.get("legend", None)
        custom_style = _custom_cycler_iterate(curve_number, 0)
        if label is not None:
            handles.append(Line2D([0], [0], label=label, **custom_style))

    if handles:
        ax.legend(handles=handles, loc=0)


def _custom_cycler_iterate(index, family_index=0, markevery=None):
    """Returns a dict with keys and values in accordance to the Matplotlib
    syntax.

    The graph style includes a color from a list defined for colour-blind people
    by https://personal.sron.nl/~pault/, depending if the data is compared with
    other data.
    """

    custom_cycler = [
        "#4477AA",
        "#66CCEE",
        "#228833",
        "#CCBB44",
        "#EE6677",
        "#AA3377",
        "#BBBBBB",
    ]
    custom_linestyles = ["solid", "dashed", "dotted", "dash", "dash", "dash", "dash"]
    custom_symbols = ["d", "o", "v", "^", ">", "<", "s"]

    index_ = index % len(custom_cycler)
    family_index_ = family_index % len(custom_linestyles)

    custom_style = {
        "color": custom_cycler[index_],
        "linestyle": custom_linestyles[family_index_],
        "marker": custom_symbols[index_],
        "markevery": markevery,
    }

    if index > len(custom_cycler):
        warnings.warn(f"More than {len(custom_cycler)} plots on one graph!")

    if family_index > len(custom_linestyles):
        warnings.warn(
            f"More than {len(custom_linestyles)} types of curves on one graph"
        )

    return custom_style


def _symbols_markevery(x):
    """compute indices where to show symbols

    Args:
        x (numpy array): x vector
    """

    if len(x) < 10:  # Failsafe for too short signals
        return list(range(len(x)))

    nb_markers = 10
    # set step between two symbols
    step = len(x) / (nb_markers - 1)
    # list of symbols indexes (first symbol at first index)
    symbols_indexes = [int(k * step) for k in range(nb_markers)]

    symbols_indexes[-1] = min(symbols_indexes[-1], len(x) - 1)

    return symbols_indexes
